(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DpFramework imports Trs Graph begin


text {* A \emph{dependency pair problem} (DP problem, for short) is a tuple consisting
of two rewrite systems *} 
types ('f,'v)dpp = "('f,'v)trs \<times> ('f,'v)trs"

text {* A \emph{graph-DP problem} is a tuple consisting of a graph (over DPs) and a rewrite system *}
types ('f,'v)gdpp = "('f,'v)rule graph \<times> ('f,'v)trs"

definition minimal_cond :: "('f,'v)trs \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> bool"
where "minimal_cond R s t \<sigma> \<equiv> \<forall> i. (SN_elt (rstep R) (t i \<cdot> \<sigma> i))"


fun min_graph_ichain ::
 "('f,'v)gdpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> bool"
where "min_graph_ichain (P,R) s t \<sigma> = (
 (\<forall>i. (s i, t i) \<in> nodes P) \<and>
 (\<forall>i. ((s i,t i),(s(Suc i),t(Suc i))) \<in> edges P) \<and>
 (\<forall>i. ((t i)\<cdot>(\<sigma> i),(s(Suc i))\<cdot>(\<sigma> (Suc i))) \<in> (rstep R)^*) \<and>
 minimal_cond R s t \<sigma>
)"

definition ideriv :: "('f,'v)trs \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> bool"
where "ideriv R t \<equiv> \<forall> i. (t i, t (Suc i)) \<in> rstep R"

fun ichain ::
 "('f,'v)dpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> bool"
where "ichain (P,R) s t \<sigma> = (
 (\<forall>i. ((s i,t i) \<in> P)) \<and> 
 (\<forall>i. ((t i)\<cdot>(\<sigma> i),(s(Suc i))\<cdot>(\<sigma> (Suc i))) \<in> (rstep R)^*) 
)"

text {* A \emph{minimal infinite chain} is an infinite chain where additionally all @{text \<sigma>}-instances
of terms in the sequence~@{text t} are terminating w.r.t.~the system~@{text R}. *}
fun min_ichain ::
 "('f,'v)dpp \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> bool"
where "min_ichain (P,R) s t \<sigma> = (ichain (P,R) s t \<sigma> \<and> minimal_cond R s t \<sigma>)"


text {* An auxiliary function to stack contexts (and substitutions) in order to reconstruct
DP steps in the original TRS. *}
fun D :: "(nat \<Rightarrow> ('f,'v)ctxt) \<Rightarrow> (nat \<Rightarrow> ('f,'v)sub) \<Rightarrow> nat \<Rightarrow> ('f,'v)ctxt"
where "D C \<sigma> 0 = \<box>"
    | "D C \<sigma> (Suc i) = (D C \<sigma> i) \<circ>c ((C i) \<cdot>c (\<sigma> i))"

lemma D_clean_sub_wf:
  fixes F::"('f\<times>nat)set" and f::"('g,'v)term \<Rightarrow> ('f,'v)term" and v::"('f,'v)term \<Rightarrow> 'v"
  defines "cs \<equiv> \<lambda>\<sigma>. case \<sigma> of Sub s \<Rightarrow> Sub(clean v F \<circ> f \<circ> s)"
  assumes "\<forall>i. C i \<in> Ctxt(F)" shows "(D C (\<lambda>i. cs(\<sigma> i)) i) \<in> Ctxt(F)"
proof (induct i)
  case 0 show ?case by auto
next
  case (Suc i)
  obtain s where s: "\<sigma> i = Sub s" by (cases "\<sigma> i") auto
  from assms have "C i \<in> Ctxt(F)" by simp
  from ctxt_clean_sub_wf[OF this,of "v" "f \<circ> s"] have "(C i \<cdot>c (cs(\<sigma> i))) \<in> Ctxt(F)"
    unfolding cs_def s o_assoc by simp
  with Suc show ?case by simp
qed

lemma [simp]: "D C \<sigma> i\<langle>C i \<cdot>c \<sigma> i\<langle>t\<rangle>\<rangle> = D C \<sigma> (Suc i)\<langle>t\<rangle>" by (cases i) simp_all

theorem clean_ichain_imp_not_SN_wfrstep:
  fixes R::"('f,'v)trs" and v::"('f,'v)term \<Rightarrow> 'v"
  defines "c' \<equiv> clean (v \<circ> unsharp) ({(Plain f,n) | f n. (f,n) \<in> funas_trs R} \<union> {(Sharp f,n) | f n. (f,n) \<in> funas_trs R})"
  defines "cs' \<sigma>' \<equiv> case \<sigma>' of Sub s \<Rightarrow> Sub(c' \<circ> s)"
  assumes chain: "ichain (DP R,ID R) (\<lambda>i. c'(s i)) (\<lambda>i. c'(t i)) (\<lambda>i. cs'(\<sigma> i))"
  shows "\<not> SN(wfrstep R)"
proof (cases "wf_trs R")
  case False
  with SN_wfrstep_imp_wf_trs show ?thesis by blast
next
  case True
  from chain have dpstep: "\<forall>i. (c'(s i),c'(t i)) \<in> DP R"
    and seq: "\<forall>i. (c'(t i) \<cdot> cs'(\<sigma> i),c'(s(Suc i)) \<cdot> cs'(\<sigma>(Suc i))) \<in> (rstep(ID R))^*"
    by auto
  let ?cu = "clean v (funas_trs R) \<circ> unsharp"
  let ?cus = "\<lambda>\<sigma>. case \<sigma> of Sub s \<Rightarrow> Sub(?cu \<circ> s)"
  have "\<forall>i.\<exists>C. C \<in> Ctxt(funas_trs R) \<and> (?cu(s i),C\<langle>?cu(t i)\<rangle>) \<in> R"
  proof
    fix i from dpstep have "(c'(s i),c'(t i)) \<in> DP R" by simp
    from DP_step_in_R[OF this,of "v"]
      show "\<exists>C. C \<in> Ctxt(funas_trs R) \<and> (?cu(s i),C\<langle>?cu(t i)\<rangle>) \<in> R"
      unfolding o_def c'_def unsharp_clean_conv[of v R,unfolded o_def] by auto
  qed
  from choice[OF this] obtain C where "\<forall>i. C i \<in> Ctxt(funas_trs R) \<and> (?cu(s i),C i\<langle>?cu(t i)\<rangle>) \<in> R" by best
  hence wf_ctxt: "\<forall>i. C i \<in> Ctxt(funas_trs R)" and R: "\<forall>i. (?cu(s i),C i\<langle>?cu(t i)\<rangle>) \<in> R" by auto
  from R have "\<forall>i. (?cu(s i) \<cdot> ?cus(\<sigma> i),C i\<langle>?cu(t i)\<rangle> \<cdot> ?cus(\<sigma> i)) \<in> CS R" using csI by best
  hence "\<forall>i. (?cu(s i) \<cdot> ?cus(\<sigma> i),C i\<langle>?cu(t i)\<rangle> \<cdot> ?cus(\<sigma> i)) \<in> rstep R"
    using ccI[where C=\<box> and r="CS R"] unfolding ctxt_apply.simps by best
  hence root_steps: "\<forall>i. (?cu(s i) \<cdot> ?cus(\<sigma> i),C i \<cdot>c ?cus(\<sigma> i)\<langle>?cu(t i) \<cdot> ?cus(\<sigma> i)\<rangle>) \<in> rstep R" (is "\<forall>i. ?R i") by simp

  let ?C = "D C (\<lambda>i. ?cus(\<sigma> i))"
  let ?s = "\<lambda>i. ?C i\<langle>?cu(s i) \<cdot> ?cus(\<sigma> i)\<rangle>"
  let ?t = "\<lambda>i. ?C(Suc i)\<langle>?cu(t i) \<cdot> ?cus(\<sigma> i)\<rangle>"
  have ctxt_steps: "\<forall>i. (?s i,?t i) \<in> rstep R"
  proof
    fix i
    from root_steps[THEN spec[where x=i]] have "(?cu(s i) \<cdot> ?cus(\<sigma> i),C i \<cdot>c ?cus(\<sigma> i)\<langle>?cu(t i) \<cdot> ?cus(\<sigma> i)\<rangle>) \<in> rstep R" .
    from rstep_imp_ctxt_rstep[OF this] have "(?C i\<langle>?cu(s i) \<cdot> ?cus(\<sigma> i)\<rangle>,?C i\<langle>C i \<cdot>c ?cus(\<sigma> i)\<langle>?cu(t i) \<cdot> ?cus(\<sigma> i)\<rangle>\<rangle>) \<in> rstep R" .
    thus "(?s i,?t i) \<in> rstep R" by simp
  qed
  from seq have aux: "\<forall>i. (unsharp(c'(t i) \<cdot> cs'(\<sigma> i)),unsharp(c'(s(Suc i)) \<cdot> cs'(\<sigma>(Suc i)))) \<in> (rstep R)^*" (is "\<forall>i. ?Q i")
    using rstepID_seq_imp_rstep_seq by best
  have "\<forall>i. (?cu(t i) \<cdot> ?cus(\<sigma> i),?cu(s(Suc i)) \<cdot> ?cus(\<sigma>(Suc i))) \<in> (rstep R)^*" (is "\<forall>i. ?P i")
  proof
    fix i
    from aux have "?Q i" by simp
    thus "?P i" unfolding c'_def cs'_def unsharp_sub_distr o_def
      unfolding unsharp_clean_conv[of v R,unfolded o_def]
      by (cases "\<sigma> i") (cases "\<sigma>(Suc i)",auto simp: unsharp_clean_conv[of v R,unfolded o_def])
  qed
  hence steps: "\<forall>i. (?t i,?s(Suc i)) \<in> (rstep R)^*" using rstep_seq_closed_ctxt by best
  have big_step: "\<forall>i. (?s i,?s(Suc i)) \<in> (rstep R)^+"
  proof
    fix i
    from ctxt_steps have "(?s i,?t i) \<in> (rstep R)" by best
    moreover from steps have "(?t i,?s(Suc i)) \<in> (rstep R)^*" by fast
    ultimately show "(?s i,?s(Suc i)) \<in> (rstep R)^+" by (rule rtrancl_into_trancl2)
  qed
  from D_clean_sub_wf[OF wf_ctxt,of v unsharp] have "\<forall>i. ?C i \<in> Ctxt(funas_trs R)" by simp
  moreover have "\<forall>i. ?cu(s i) \<cdot> ?cus(\<sigma> i) \<in> Term(funas_trs R)"
  proof
    fix i
    obtain sub where sub: "\<sigma> i = Sub sub" by (cases "\<sigma> i") auto
    have "?cu(s i) \<in> Term(funas_trs R)" by auto
    from clean_sub_apply_wf[OF this,of v "unsharp \<circ> sub"] show "?cu(s i) \<cdot> ?cus(\<sigma> i) \<in> Term(funas_trs R)" unfolding sub o_assoc sub.cases .
  qed
  ultimately have "\<forall>i. ?s i \<in> Term(funas_trs R)" using ctxt_apply_wf by best
  with big_step have seq: "\<forall>i. (?s i,?s(Suc i)) \<in> {(x,y) | x y. (x,y) \<in> (rstep R)^+ \<and> x \<in> Term(funas_trs R) \<and> y \<in> Term(funas_trs R)}"
    (is "\<forall>i. _ \<in> ?steps") by blast
  have"?s 0 = ?s 0" by simp
  with seq have "?s 0 = ?s 0 \<and> (\<forall>i. (?s i,?s(Suc i)) \<in> ?steps)" by best
  hence "\<exists>S. S 0 = ?s 0 \<and> (\<forall>i. (S i,S(Suc i)) \<in> ?steps)" by best
  hence "\<exists>x S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> ?steps)" by (rule exI[where x="?s 0"])
  hence "\<not> SN ?steps" unfolding SN_defs by blast
  hence "\<not> SN((wfrstep R)^+)" unfolding wfrstep_conv[OF True] .
  with SN_trancl_SN_conv show ?thesis by auto
qed

hide const D

(*
lemma inj_clean_imp_eq_sub:
  assumes inj: "inj v" shows "clean v F s \<cdot> (Sub(inv v)) = s \<or> clean v F s = s"
proof (induct s)
  case (Var x)
  thus ?case by simp
next
 case (Fun f ss)
 hence "\<forall>s\<in>set ss.\<exists>\<tau>. clean v F s \<cdot> \<tau> = s" by best
 from bchoice[OF this] obtain sig where IH: "\<forall>s\<in>set ss. clean v F s \<cdot> sig s = s" by best
 show ?case
 proof (cases "(f,length ss) \<in> F")
   case True
   hence "clean v F (Fun f ss) = Fun f ss"
thus ?thesis using Fun by auto

lemma rstep_imp_clean_rstep_or_Id_sub:
  fixes v::"('f,'v)term \<Rightarrow> 'v" and F::"('f\<times>nat)set"
  defines [simp]: "c \<equiv> clean v F"
  assumes inj: "inj v" and subset: "funas_trs R \<subseteq> F" and step: "(s,t) \<in> rstep R"
  shows "\<exists>\<sigma>. (c s,c t \<cdot> \<sigma>) \<in> rstep R \<union> Id"
proof -
  from rstep_imp_C_s_r[OF step] obtain C l r \<sigma>
    where R: "(l,r) \<in> R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>"
    and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
  from s t show ?thesis
  proof (induct C arbitrary: s t)
    case Hole
    let ?c = "\<lambda>\<sigma>. case \<sigma> of Sub s \<Rightarrow> Sub(c \<circ> s)"
    from clean_sub_apply[OF lhs_wf[OF R subset]] have s: "c(s) = l\<cdot>?c(\<sigma>)" unfolding Hole by (cases \<sigma>) simp_all
    from clean_sub_apply[OF rhs_wf[OF R subset]] have t: "c(t) = r\<cdot>?c(\<sigma>)" unfolding Hole by (cases \<sigma>) simp_all
    from R have "(c(s),c(t)) \<in> CS R" unfolding s t by auto
    hence "(c(s),c(t)) \<in> rstep R" using ccI[where C=\<box>] by simp
    hence "(c(s),c(t) \<cdot> \<iota>) \<in> rstep R \<union> Id" by simp
    thus ?case by best
  next
    case (More f ss1 D ss2)
    then obtain \<tau> where IH: "(c(D\<langle>l\<cdot>\<sigma>\<rangle>),c(D\<langle>r\<cdot>\<sigma>\<rangle>) \<cdot> \<tau>) \<in> rstep R \<union> Id" by best
    show ?case
    proof (cases "(f,Suc(length(ss1@ss2))) \<in> F")
      case False
      with More have "c(s) = Var(v s)" by simp
      moreover from False More have "c(t) = Var(v t)" by simp
      ultimately have "(c(s),c(t)) \<in> Id" by simp
      thus ?thesis ..
    next
      case True
      let ?s = "c(D\<langle>l\<cdot>\<sigma>\<rangle>)"
      let ?t = "c(D\<langle>r\<cdot>\<sigma>\<rangle>)"
      from IH show ?thesis
      proof
        assume "(?s,?t) \<in> Id" with True More show ?thesis by auto
      next
        let ?C = "More f (map c ss1) \<box> (map c ss2)"
        assume "(?s,?t) \<in> rstep R"
        from rstep_imp_ctxt_rstep[OF this,where C="?C"]
          have "(c(s),c(t)) \<in> rstep R" unfolding More using True by simp
        thus ?thesis ..
      qed
    qed
  qed
qed
*)

lemma ichain_imp_clean_ichain:
  fixes P R::"('f,'v)trs" and F::"('f\<times>nat)set"
  defines "c \<equiv> clean (\<lambda>x.(SOME(x::'v). True)) F"
  defines "cs \<equiv> (\<lambda>\<sigma>. case \<sigma> of Sub s \<Rightarrow> Sub(c \<circ> s))"
  assumes subset: "funas_trs P \<union> funas_trs R \<subseteq> F"
    and ichain: "ichain (P,R) s t \<sigma>"
  shows "ichain (P,R) (\<lambda>i. c(s i)) (\<lambda>i. c(t i)) (\<lambda>i. cs(\<sigma> i))"
proof -
  from ichain have dpstep: "\<forall>i. (s i,t i) \<in> P"
    and seq: "\<forall>i. (t i \<cdot> \<sigma> i,s(Suc i)\<cdot>\<sigma>(Suc i)) \<in> (rstep R)^*" by auto
  from subset have subset_P: "funas_trs P \<subseteq> F" by simp
  from subset have subset_R: "funas_trs R \<subseteq> F" by simp
  have wf_s: "\<forall>i. s i \<in> Term(F)"
  proof
    fix i
    from dpstep have "(s i,t i) \<in> P" by simp
    from lhs_wf[OF this subset_P] show "s i \<in> Term(F)" .
  qed
  have wf_t: "\<forall>i. t i \<in> Term(F)"
  proof
    fix i
    from dpstep have "(s i,t i) \<in> P" by simp
    from rhs_wf[OF this subset_P] show "t i \<in> Term(F)" .
  qed
  have c_P: "\<forall>i. (c(s i),c(t i)) \<in> P"
  proof
    fix i
    from dpstep have P: "(s i,t i) \<in> P" by simp
    from lhs_wf[OF P subset_P] have "s i \<in> Term(F)" .
    hence s: "c(s i) = s i" using clean_fun_terms_apply_id by (simp only: c_def)
    from rhs_wf[OF P subset_P] have "t i \<in> Term(F)" .
    hence t: "c(t i) = t i" using clean_fun_terms_apply_id by (simp only: c_def)
    from P show "(c(s i),c(t i)) \<in> P" unfolding s t .
  qed
  have seq': "\<forall>i. (c(t i\<cdot>(\<sigma> i)),c(s(Suc i)\<cdot>(\<sigma>(Suc i)))) \<in> (rstep R)^*" (is "\<forall>i. (c(?t i),c(?s i)) \<in> _")
  proof
    fix i
    from seq have "(?t i,?s i) \<in> (rstep R)^*" by simp
    thus "(c(?t i),c(?s i)) \<in> (rstep R)^*"
    proof (induct rule: rtrancl_induct)
      case base show ?case by simp
    next
      case (step y z)
      from subset have "funas_trs R \<subseteq> F" by auto
      from rstep_imp_clean_rstep_or_Id[OF this `(y,z) \<in> rstep R`]
        have "(c(y),c(z)) \<in> rstep R \<union> Id" by (simp only: c_def)
      thus ?case
      proof
        assume "(c y,c z) \<in> Id" with step show ?thesis by simp
      next
        assume "(c y,c z) \<in> rstep R" with step show ?thesis by auto
      qed
    qed
  qed
  have "\<forall>i. (c(t i) \<cdot> cs(\<sigma> i),c(s(Suc i)) \<cdot> cs(\<sigma>(Suc i))) \<in> (rstep R)^*"
  proof
    fix i
    obtain \<sigma>' where sub_i: "\<sigma> i = Sub \<sigma>'" by (cases "\<sigma> i") auto
    from wf_t have wf_t: "t i \<in> Term(F)" by simp
    from clean_sub_apply[OF this,of "\<lambda>x.(SOME(x::'v). True)" "\<sigma>'"]
      have fst: "c(?t i) = c(t i) \<cdot> cs(\<sigma> i)"
      unfolding c_def cs_def sub_i using wf_t by simp
    obtain \<sigma>'' where sub_si: "\<sigma>(Suc i) = Sub \<sigma>''" by (cases "\<sigma>(Suc i)") auto
    from wf_s have wf_s: "s(Suc i) \<in> Term(F)" by simp
    from clean_sub_apply[OF this,of "\<lambda>x.(SOME(x::'v). True)" "\<sigma>''"]
      have snd: "c(?s i) = c(s(Suc i)) \<cdot> cs(\<sigma>(Suc i))"
      unfolding c_def cs_def sub_si using wf_s by simp
    from seq' have "(c(?t i),c(?s i)) \<in> (rstep R)^*" by simp
    thus "(c(t i) \<cdot> cs(\<sigma> i),c(s(Suc i)) \<cdot> cs(\<sigma>(Suc i))) \<in> (rstep R)^*" unfolding fst snd .
  qed
  with c_P show ?thesis by auto
qed


lemma min_ichain_imp_ichain: assumes "min_ichain DPP s t \<sigma>" shows "ichain DPP s t \<sigma>"
using assms
by (cases DPP, auto)

fun shift :: "(nat \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> (nat \<Rightarrow> 'a)"
  where "shift f j = (\<lambda> i. f (i+j))"

definition ci_subset :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool" 
where "ci_subset R S \<equiv> \<forall> l r. (l,r) \<in> R \<longrightarrow> (\<exists> l' r' C. (l',r') \<in> S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>)"

notation (xsymbols)
  ci_subset ("(_/ \<subseteq>ci _)" [50,51] 50)

lemma ci_subsetI : assumes "R \<subseteq> S"
  shows "ci_subset R S"
unfolding ci_subset_def
proof (intro allI impI)
  fix l r
  assume "(l,r) \<in> R"
  with assms have S: "(l,r) \<in> S" by auto
  show "\<exists> l' r' C. (l',r') \<in> S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>"
    by (rule exI[of _ l], rule exI[of _ r], rule exI[of _ \<box>], simp add: S)
qed

lemma ci_subset_refl: "R \<subseteq>ci R"
proof (unfold ci_subset_def, intro allI impI)
  fix l r
  assume rule: "(l,r) \<in> R"
  show "\<exists> l' r' C.  (l', r') \<in> R \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>"
    by (rule exI[of _ l], rule exI[of _ r], rule exI[of _ \<box>], auto simp: rule)
qed
  
lemma rstep_ci_mono: assumes "R \<subseteq>ci S"
  shows "rstep R \<subseteq> rstep S"
proof (simp only: rstep_subset_characterization, intro allI impI)
  fix l r
  assume "(l,r) \<in> R"
  with assms obtain l' r' C where cond: "(l',r') \<in> S \<and> l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>" unfolding ci_subset_def by blast
  show "\<exists> l' r' C \<sigma>. (l',r') \<in> S \<and> l = C\<langle>l' \<cdot> \<sigma>\<rangle> \<and> r = C\<langle>r' \<cdot> \<sigma>\<rangle>"
    by (rule exI[of _ l'], rule exI[of _ r'], rule exI[of _ C], rule exI[of _ \<iota>], simp add: cond)
qed 

lemma minimal_cond_mono : 
  assumes subset: "R \<subseteq>ci R'" and cond: "minimal_cond R' s t \<sigma>" 
  shows "minimal_cond R s t \<sigma>"
unfolding minimal_cond_def
proof 
  fix i
  from cond have SN: "SN_elt (rstep R') (t i \<cdot> \<sigma> i)"
    unfolding minimal_cond_def by auto
  have "rstep R \<subseteq> rstep R'" by (rule rstep_ci_mono, rule subset)
  with SN show "SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by (rule SN_imp_SN_subset)
qed

lemma min_ichainI[intro] : 
  assumes "R \<subseteq>ci R'" and "minimal_cond R' s t \<sigma>" and "ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)" 
  shows "min_ichain (P,R) (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  from assms have "minimal_cond R s t \<sigma>" using minimal_cond_mono by auto
  hence "minimal_cond R (shift s i) (shift t i) (shift \<sigma> i)" 
    unfolding minimal_cond_def by auto
  with assms show ?thesis by auto
qed


fun initial_dpp :: "('f,'v)trs \<Rightarrow> ('f shp,'v)dpp"
where "initial_dpp R = (DP R, ID R)"

definition finite_dpp :: "('f,'v)dpp \<Rightarrow> bool" where "finite_dpp DPP = (\<not>(\<exists>s t \<sigma>. min_ichain DPP s t \<sigma>))"

definition finite_gdpp :: "('f,'v)gdpp \<Rightarrow> bool" where "finite_gdpp DPP = (\<not>(\<exists>s t \<sigma>. min_graph_ichain DPP s t \<sigma>))"

fun next4 :: "('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow>
 (('f shp,'v)term \<times> ('f shp,'v)term \<times> ('f,'v)term \<times> ('f shp,'v)sub)"
where "next4 R s = (SOME (l,r,u,\<sigma>).
 (l,r) \<in> DP R \<and> r\<cdot>\<sigma> = sharp u \<and> Tinf R u \<and> (sharp s,l\<cdot>\<sigma>) \<in> (rstep(ID R))^*)"

fun concrete_seq :: "('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> nat \<Rightarrow>
 (('f,'v)term \<times> ('f shp,'v)term \<times> ('f shp,'v)term \<times> ('f,'v)term \<times> ('f shp,'v)sub)"
where "concrete_seq R s 0       = (s,next4 R s)"
    | "concrete_seq R r (Suc n) = (let (_,_,_,s,_) = concrete_seq R r n in (s,next4 R s))"

lemma someI_ex4: "\<exists>w x y z. P(w,x,y,z) \<Longrightarrow> P(SOME(w,x,y,z). P(w,x,y,z))" using some_eq_ex by force

theorem not_SN_imp_min_ichain:
 fixes R :: "('f,'v)trs"
 assumes "wf_trs R" and "\<not> SN(rstep R)"
 shows "\<exists>s t \<sigma>. min_ichain (initial_dpp R) s t \<sigma>"
proof -
 from `\<not>(SN(rstep R))` obtain s' :: "('f,'v)term" where "\<not>(SN_elt (rstep R) s')"
   unfolding SN_def by auto
 then obtain s where "s' \<unrhd>  s" and "Tinf R s" using not_SN_imp_subt_Tinf by auto
 let ?I = "\<lambda>i. fst(concrete_seq R s i)"
 let ?S = "\<lambda>i. fst(snd(concrete_seq R s i))"
 let ?T = "\<lambda>i. fst(snd(snd(concrete_seq R s i)))"
 let ?\<sigma> = "\<lambda>i. snd(snd(snd(snd(concrete_seq R s i))))"
 let ?P = "\<lambda>s l r u \<sigma>. (l,r) \<in> DP R \<and> r\<cdot>\<sigma> = sharp u \<and> Tinf R u \<and> (sharp s,l\<cdot>\<sigma>) \<in> (rstep(ID R))^*"
 let ?P' =  "\<lambda>(l,r,u,\<sigma>). ?P s l r u \<sigma>"
 have main: "\<forall>i. ?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
 proof
  fix i show "?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
  proof (induct i)
   case 0 
   have "\<exists>l r u \<sigma>. ?P'(l,r,u,\<sigma>)" using Tinf_lemma_DP_explicit_sub[OF `wf_trs R` `Tinf R s`] by auto
   hence "?P'(next4 R s)" unfolding next4.simps using someI_ex4[where P = "?P'"] by best
   have ini: "?I 0 = s" unfolding concrete_seq.simps by simp
   have s: "?S 0 = fst(next4 R s)" unfolding concrete_seq.simps by simp
   have t: "?T 0 = fst(snd(next4 R s))" unfolding concrete_seq.simps by simp
   have sub: "?\<sigma> 0 = snd(snd(snd(next4 R s)))" unfolding concrete_seq.simps Let_def split_def by auto
   have nxt: "?I(Suc 0) = fst(snd(snd(next4 R s)))" unfolding concrete_seq.simps Let_def split_def by auto
   show ?case unfolding ini s t sub nxt using `?P'(next4 R s)` by auto
  next
   case (Suc i)
   let ?P'' = "\<lambda>(w,x,y,z). ?P (?I(Suc i)) w x y z"
   from Suc have "Tinf R (?I(Suc i))" by simp
   have "\<exists>l r u \<sigma>. ?P''(l,r,u,\<sigma>)" using Tinf_lemma_DP_explicit_sub[OF `wf_trs R` `Tinf R (?I(Suc i))`] by auto
   hence IH: "?P''(next4 R (?I(Suc i)))" unfolding next4.simps using someI_ex4[where P = "?P''"] by simp
   have s: "?S(Suc i) = fst(next4 R (?I(Suc i)))" unfolding concrete_seq.simps Let_def split_def by simp
   have t: "?T(Suc i) = fst(snd(next4 R (?I(Suc i))))" unfolding concrete_seq.simps Let_def split_def by simp
   have sub: "?\<sigma>(Suc i) = snd(snd(snd(next4 R (?I(Suc i)))))" unfolding concrete_seq.simps Let_def split_def by simp
   have nxt: "?I(Suc(Suc i)) = fst(snd(snd(next4 R (?I(Suc i)))))" unfolding concrete_seq.simps Let_def split_def by simp
   show ?case unfolding s t sub nxt using IH by auto
  qed
 qed
 from main have P: "\<forall>i. (?S i,?T i) \<in> DP R" by simp
 from main have eq: "\<And>i. sharp(?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by auto
 have R: "\<forall>i. (?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> (rstep(ID R))^*"
 proof
  fix i show "(?T i \<cdot> ?\<sigma> i,?S(Suc i)\<cdot>?\<sigma>(Suc i)) \<in> (rstep(ID R))^*"
  proof (cases i)
   case 0
   from main have first: "?T 0 \<cdot> ?\<sigma> 0 = sharp(?I(Suc 0))" by best
   from main have "(sharp(?I(Suc 0)),?S(Suc 0)\<cdot>?\<sigma>(Suc 0)) \<in> (rstep(ID R))^*" by best
   thus ?thesis unfolding 0 unfolding first by simp
  next
   case (Suc i) show ?thesis unfolding eq[symmetric] using main by best
  qed
 qed
 from P and R have ichain: "ichain (initial_dpp R) ?S ?T ?\<sigma>" by auto
 have Tinf: "\<forall>i. Tinf R (?I i)"
 proof
  fix i show "Tinf R (?I i)" using main by (induct i) (auto simp: `Tinf R s`)
 qed
 have SN_t: "\<forall>i. SN_elt (rstep(ID R)) (?T i \<cdot> ?\<sigma> i)"
 proof
  fix i
  from Tinf have "Tinf R (?I(Suc i))" by best
  from eq have eq': "sharp(?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by simp
  from Tinf_imp_SN_sharp[OF `wf_trs R` `Tinf R (?I(Suc i))`]
  show "SN_elt (rstep(ID R)) (?T i \<cdot> ?\<sigma> i)" unfolding eq' by simp
 qed
 have SN_s: "\<forall>i. SN_elt (rstep(ID R)) (?S i \<cdot> ?\<sigma> i)"
 proof
  fix i show "SN_elt (rstep(ID R)) (?S i \<cdot> ?\<sigma> i)"
  proof (cases i)
   case 0
   have "?I 0 = s" by simp
   have SN_sharp: "SN_elt (rstep(ID R)) (sharp s)" using Tinf_imp_SN_sharp[OF `wf_trs R` `Tinf R s`] by simp
   from main have "(sharp(?I 0),?S 0 \<cdot> ?\<sigma> 0) \<in> (rstep(ID R))^*" by best
   hence "(sharp s,?S 0 \<cdot> ?\<sigma> 0) \<in> (rstep(ID R))^*" by simp
   from rtrancl_step_preserves_SN[OF this SN_sharp] show ?thesis unfolding 0 by simp
  next
   case (Suc i)
   from SN_t have SN: "SN_elt (rstep(ID R)) (?T i \<cdot> ?\<sigma> i)" by simp
   from R have "(?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> (rstep(ID R))^*" by simp
   from rtrancl_step_preserves_SN[OF this SN] show ?thesis unfolding Suc by simp
  qed
 qed
 from ichain and SN_s and SN_t
 show ?thesis unfolding initial_dpp.simps min_ichain.simps minimal_cond_def by blast
qed
 
hide const concrete_seq next4
no_notation Plain ("_@")

lemma min_graph_ichain_imp_min_ichain : assumes "min_graph_ichain (P,R) s t \<sigma>"
  shows "min_ichain (nodes P, R) s t \<sigma>"
using assms
by (simp add: nodes_def edges_def)

lemma gdpp_to_dpp : assumes fin: "finite_dpp (nodes P, R)" shows "finite_gdpp (P, R)"
  using assms min_graph_ichain_imp_min_ichain  
  by (unfold finite_dpp_def finite_gdpp_def, blast)
    
lemma dpp_lift_to_gdpp : fixes P :: "('f, 'v)rule graph"
  assumes 
  step: "\<And> s t \<sigma>. min_ichain (nodes P, R) s t \<sigma> \<Longrightarrow> \<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
  and res: "finite_gdpp (node_restriction P rP, R)" (is "finite_gdpp (?rP, _)")
  shows "finite_gdpp (P, R)"
proof (unfold finite_gdpp_def, rule notI, clarify)
  fix s t \<sigma>
  assume chain: "min_graph_ichain (P, R) s t \<sigma>"
  hence "min_ichain (nodes P, R) s t \<sigma>" by (rule min_graph_ichain_imp_min_ichain)
  with step have "\<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by simp
  from this obtain j where rPchain: "min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
  with chain have "min_graph_ichain (?rP, R) (shift s j) (shift t j) (shift \<sigma> j)" 
  proof (simp add: node_restriction_def edges_def nodes_def, clarify)
    fix i
    assume "\<forall> i. (s (i+j), t (i + j)) \<in> rP"
    hence "(s (Suc i + j), t (Suc i + j)) \<in> rP" ..
    thus "(s (Suc (i+j)), t (Suc (i+j))) \<in> rP" by simp
  qed
  with res show False by (unfold finite_gdpp_def, blast)
qed

lemma finite_ini_imp_SN:
  assumes "wf_trs R" and "finite_dpp(initial_dpp R)" shows "SN(rstep R)"
proof (rule ccontr)
 assume "\<not>SN(rstep R)"
 from not_SN_imp_min_ichain[OF `wf_trs R` `\<not>(SN(rstep R))`]
 obtain s t \<sigma> where "min_ichain (initial_dpp R) s t \<sigma>" by best
 with `finite_dpp(initial_dpp R)` show False unfolding finite_dpp_def by best
qed

lemma SN_wfrstep_imp_SN_rstep:
  assumes wf: "wf_trs R" and SN: "SN(wfrstep R)" shows "SN(rstep R)"
proof (rule ccontr)
  let ?F = "{(Plain f,n) | f n. (f,n) \<in> funas_trs R} \<union> {(Sharp f,n) | f n. (f,n) \<in> funas_trs R}"
  let ?c = "clean (\<lambda>x.(SOME(x::'v). True)) ?F"
  let ?cs = "\<lambda>\<sigma>. case \<sigma> of Sub s \<Rightarrow> Sub(?c \<circ> s)"
  assume "\<not> ?thesis"
  with finite_ini_imp_SN[OF wf] obtain s t \<sigma> where ichain: "ichain (DP R, ID R) s t \<sigma>" unfolding finite_dpp_def by simp blast
  hence clean_ichain: "ichain (DP R,ID R) (\<lambda>i. ?c(s i)) (\<lambda>i. ?c(t i)) (\<lambda>i. ?cs(\<sigma> i))" 
    by (rule ichain_imp_clean_ichain[OF funas_DP_ID_subset])
  hence "\<not> SN(wfrstep R)" using clean_ichain_imp_not_SN_wfrstep[of R "\<lambda>x.(SOME x. True)"] unfolding o_def by blast
  with SN show False by blast
qed
  

lemma SN_wfrstep_SN_rstep_conv:
  fixes R::"('f,'v)trs" shows "SN(wfrstep R) = SN(rstep R)"
proof (cases "wf_trs R")
  case True
  {
    assume "\<not> SN(wfrstep R)"
    then obtain S where "\<forall>i. (S i,S(Suc i)) \<in> wfrstep R" unfolding SN_defs by auto
    hence "\<not> SN(rstep R)" unfolding SN_defs wfrstep_def by blast
  }
  with SN_wfrstep_imp_SN_rstep[OF True] show ?thesis by auto
next
  case False thus ?thesis using SN_rstep_imp_wf_trs SN_wfrstep_imp_wf_trs by auto
qed

lemma SN_imp_finite_ini:
  assumes wf: "wf_trs R" and SN: "SN(rstep R)" shows "finite_dpp(initial_dpp R)"
proof (rule ccontr)
  assume "\<not> finite_dpp(initial_dpp R)"
  then obtain s t \<sigma> where ichain: "ichain (DP R,ID R) s t \<sigma>" unfolding finite_dpp_def by auto
  have "\<not> SN(wfrstep R)" using
    ichain_imp_clean_ichain[OF funas_DP_ID_subset ichain] 
    clean_ichain_imp_not_SN_wfrstep[of R "\<lambda>x.(SOME x. True)"] unfolding o_def by blast
  thus False using SN by (simp add: SN_wfrstep_SN_rstep_conv)
qed

text {* Well-formedness is important since a variable lhs implies that @{text R} is not terminating. Still,
there is no \emph{minimal} infinite chain, since there is no term that is strongly normalizing w.r.t.\ to
@{text R}. *}
lemma SN_iff_finite_ini: assumes "wf_trs R" shows "SN(rstep R) = finite_dpp(initial_dpp R)"
using SN_imp_finite_ini[OF assms] finite_ini_imp_SN[OF assms] by blast

(* and a version which preprocesses the rules *)
fun computeRuleMap :: "('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL) option"
where "computeRuleMap [] = Some (\<lambda> f. [])"
    | "computeRuleMap ((Fun f ts,r) # rules) = (case computeRuleMap rules of None \<Rightarrow> None | Some rm \<Rightarrow> 
            Some (rm(f := (Fun f ts,r) # (rm f))))"
    | "computeRuleMap ((Var _, _) # rules) = None"

lemma computeRuleMapSound : assumes "computeRuleMap R = Some rm"
  shows "(set (rm f)) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f}"
using assms 
proof (induct R arbitrary: rm f, force)
  case (Cons rule rules)
  show ?case
  proof (cases rule)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var v)
      with Cons Pair show ?thesis by simp
    next
      case (Fun g ts)
      with Cons Pair show ?thesis by (cases "computeRuleMap rules", simp, cases "f = g", auto)
    qed
  qed
qed

lemma computeRuleMapSound2 : "(computeRuleMap R = None) = (\<exists> (l, r) \<in> set R. root l = None)"
using assms 
proof (induct R, simp)
  case (Cons rule rules)
  show ?case 
  proof (cases rule)
    case (Pair l r)
    with Cons show ?thesis by (cases l, simp, cases "computeRuleMap rules", auto)
  qed
qed


definition sound_tt :: "(('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool) \<Rightarrow> bool"
  where "sound_tt tt \<equiv> \<forall> R rR. tt R rR \<longrightarrow> (SN(rstep rR) \<longrightarrow> SN(rstep R))"

definition ci_tt :: "(('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool) \<Rightarrow> bool"
  where "ci_tt tt \<equiv> \<forall> R rR. tt R rR \<longrightarrow> (rR \<subseteq>ci R \<and> (\<forall> t. ideriv R t \<longrightarrow> (\<exists> i. ideriv rR (shift t i))))"

definition sound_tt_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_tt_impl tt \<equiv> \<forall> R rR. tt R rR \<longrightarrow> (SN(rstep (set rR)) \<longrightarrow> SN(rstep (set R)))"

definition ci_tt_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "ci_tt_impl tt \<equiv> \<forall> R rR. tt R rR \<longrightarrow> (set rR \<subseteq>ci set R \<and> (\<forall> t. ideriv (set R) t \<longrightarrow> (\<exists> i. ideriv (set rR) (shift t i))))"

definition chain_identifying_proc :: "(('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool) \<Rightarrow> bool"
  where "chain_identifying_proc proc \<equiv> \<forall> P R rP rR. (proc P R rP rR \<longrightarrow> (rR \<subseteq>ci R \<and> (\<forall> s t \<sigma> . min_ichain (P,R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)))))"

definition sound_proc :: "(('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool) \<Rightarrow> bool"
  where "sound_proc proc \<equiv> \<forall> P R rP rR. (proc P R rP rR \<longrightarrow> (finite_dpp (rP, rR) \<longrightarrow> finite_dpp (P, R)))"

lemma ci_tt_to_sound_tt[simp] : 
  assumes ci_tt: "ci_tt tt"
  shows "sound_tt tt"
unfolding sound_tt_def
proof (intro allI, intro impI)
  fix R rR
  assume tt: "tt R rR" and SN: "SN(rstep rR)"  
  show "SN(rstep R)"
    unfolding SN_def SN_elt_def
  proof (rule allI, rule notI)
    fix t0
    assume "\<exists> t. t 0 = t0 \<and> (\<forall> i. (t i, t (Suc i)) \<in> rstep R)"
    from this obtain t where "t 0 = t0 \<and> (\<forall> i. (t i, t (Suc i)) \<in> rstep R)" ..
    hence "ideriv R t" unfolding ideriv_def by auto
    with ci_tt tt ci_tt_def[of tt] obtain i where "ideriv rR (shift t i)" by force
    hence "\<exists> x. (shift t i) 0 = x \<and> (\<forall> j. (shift t i j, shift t i (Suc j)) \<in> rstep rR)" 
      unfolding ideriv_def by auto
    thus False using SN unfolding SN_def SN_elt_def by blast
  qed
qed

lemma ci_proc_to_sound_proc[simp] : 
  assumes ci_proc: "chain_identifying_proc proc"
  shows "sound_proc proc"
unfolding sound_proc_def
proof (intro allI, intro impI)
  fix P R rP rR
  assume cond: "proc P R rP rR" and finite: "finite_dpp (rP, rR)"
  show "finite_dpp (P,R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (P,R)"
    from this obtain s t \<sigma> where min: "min_ichain (P,R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where subset: "rR \<subseteq>ci R" and ichain: "ichain (rP,rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_proc_def by force
    have "min_ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = R and R = rR and P = rP] using ichain min subset by auto
    hence "\<not> finite_dpp (rP, rR)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

definition chain_identifying_proc_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "chain_identifying_proc_impl proc \<equiv> \<forall> P R rP rR. (proc P R (computeRuleMap R) rP rR \<longrightarrow> (set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))))"

definition sound_proc_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_proc_impl proc \<equiv> \<forall> P R rP rR. (proc P R (computeRuleMap R) rP rR \<longrightarrow> (finite_dpp (set rP, set rR) \<longrightarrow> finite_dpp (set P, set R)))"

definition reverse_trs :: "('f,'v)trsL \<Rightarrow> ('f,'v)trsL"
where "reverse_trs r \<equiv> map (\<lambda> lr. (snd lr, fst lr)) r"

definition chain_identifying_onlyDP_proc_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "chain_identifying_onlyDP_proc_impl proc \<equiv> \<forall> P R rP. (proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP \<longrightarrow> ((\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)))))"

definition sound_onlyDP_proc_impl :: "(('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool) \<Rightarrow> bool"
where "sound_onlyDP_proc_impl proc \<equiv> \<forall> P R rP. (proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP \<longrightarrow> (finite_dpp (set rP, set R) \<longrightarrow> finite_dpp (set P, set R)))"

lemma ci_onlyDP_imp_sound_onlyDP[simp] : 
  assumes ci_proc: "chain_identifying_onlyDP_proc_impl proc"
  shows "sound_onlyDP_proc_impl proc" 
unfolding sound_onlyDP_proc_impl_def
proof ((rule allI)+, (rule impI)+)
  fix P R rP 
  assume cond: "proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP" and finite: "finite_dpp (set rP, set R)"
  show "finite_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (set P, set R)"
    from this obtain s t \<sigma> where min: "min_ichain (set P, set R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where ichain: "ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_onlyDP_proc_impl_def by blast
    have "min_ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = "set R" and R = "set R" and P = "set rP"] using ichain min ci_subset_refl[of "set R"] by auto
    hence "\<not> finite_dpp (set rP, set R)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

lemma ci_tt_impl_to_sound_tt[simp] : 
  assumes ci_tt: "ci_tt_impl tt"
  shows "sound_tt_impl tt"
unfolding sound_tt_impl_def
proof (intro allI, intro impI)
  fix R rR
  assume tt: "tt R rR" and SN: "SN(rstep (set rR))"  
  show "SN(rstep (set R))"
    unfolding SN_def SN_elt_def
  proof (rule allI, rule notI)
    fix t0
    assume "\<exists> t. t 0 = t0 \<and> (\<forall> i. (t i, t (Suc i)) \<in> rstep (set R))"
    from this obtain t where "t 0 = t0 \<and> (\<forall> i. (t i, t (Suc i)) \<in> rstep (set R))" ..
    hence "ideriv (set R) t" unfolding ideriv_def by auto
    with ci_tt tt ci_tt_impl_def[of tt] obtain i where "ideriv (set rR) (shift t i)" by force
    hence "\<exists> x. (shift t i) 0 = x \<and> (\<forall> j. (shift t i j, shift t i (Suc j)) \<in> rstep (set rR))" 
      unfolding ideriv_def by auto
    thus False using SN unfolding SN_def SN_elt_def by blast
  qed
qed

lemma apply_ci_onlyDP_proc : assumes ci: "chain_identifying_onlyDP_proc_impl proc"
  and fin: "finite_dpp (set rP, set R)"
  and proc: "proc P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) rP"
  shows "finite_dpp (set P, set R)"
proof -
  from ci have "sound_onlyDP_proc_impl proc" by simp
  thus ?thesis unfolding sound_onlyDP_proc_impl_def using fin proc by blast
qed

lemma ci_imp_sound[simp] : 
  assumes ci_proc: "chain_identifying_proc_impl proc"
  shows "sound_proc_impl proc" 
unfolding sound_proc_impl_def
proof ((rule allI)+, (rule impI)+)
  fix P R rP rR
  assume cond: "proc P R (computeRuleMap R) rP rR" and finite: "finite_dpp (set rP, set rR)"
  show "finite_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (set P, set R)"
    from this obtain s t \<sigma> where min: "min_ichain (set P, set R) s t \<sigma>" unfolding finite_dpp_def by auto
    from this cond ci_proc obtain i where ichain: "set rR \<subseteq>ci set R \<and> ichain (set rP,set rR) (shift s i) (shift t i) (shift \<sigma> i)" unfolding chain_identifying_proc_impl_def by blast
    have "min_ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)" 
      using min_ichainI[where R' = "set R" and R = "set rR" and P = "set rP"] using ichain min by auto
    hence "\<not> finite_dpp (set rP, set rR)" unfolding finite_dpp_def by blast
    with finite show False by auto
  qed
qed

lemma apply_ci_proc : assumes ci: "chain_identifying_proc_impl proc"
  and fin: "finite_dpp (set rP, set rR)"
  and proc: "proc P R (computeRuleMap R) rP rR"
  shows "finite_dpp (set P, set R)"
proof -
  from ci have "sound_proc_impl proc" by simp
  thus ?thesis unfolding sound_proc_impl_def using fin proc by blast
qed

lemma apply_ci_tt : assumes ci: "ci_tt_impl tt"
  and fin: "SN(rstep (set rR))"
  and tt: "tt R rR"
  shows "SN(rstep (set R))"
proof -
  from ci have "sound_tt_impl tt" by simp
  thus ?thesis unfolding sound_tt_impl_def using fin tt by blast
qed

definition internal_error where "internal_error \<equiv> (
  ''internal error'' +#+ shows_nl +@+
  ''please contact one of the CeTA members'' +#+ shows_nl +@+
  ''http://cl-informatik.uibk.ac.at/software/ceta'' +#+ shows_nl +@+
  ''ceta@informatik.uibk.ac.at'' +#+ shows_nl
)"


end
